/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

/*
 * Created on 16.05.2003
 */
package at.spardat.xma.boot.transport;

import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

import at.spardat.xma.boot.Statics;
import at.spardat.xma.boot.logger.LogLevel;
import at.spardat.xma.boot.logger.Logger;

/**
 * custom URI class for xma usage. This abstraction of URI brings the capability to
 * use the same URI even for different implementations of {@link Transport}.
 * At this time only a mapping to HTTP-URLs is implemented but additional protocols
 * may be added in the future.
 * <br><br>
 * The URIs used by XMA contain of the following parts:<br><code>
 * &lt;protocol&gt;://&lt;server&gt;[_&lt;environment&gt;][_&lt;installation&gt;][:&lt;port&gt;]/&lt;application&gt;[/&lt;component&gt;][/&lt;resource&gt;][&&lt;parameters&gt;]
 * </code><ul>
 * <li><b>protocol</b> the protocol to use. currently only <code>http</code> is supported.
 * <li><b>server</b> the DNS-name of the server hosting the xma application to reference.
 * <li>optional <b>envrionment</b> the envrionment specific part of the DNS-name of the server (not implemented yet)
 * <li>optional <b>installation</b> the part of the DNS-name specific for the organisation operating the
 *      server in a given location (not implemented yet)
 * <li>optional <b>port</b> the port used by the protocol running on the given server. Can be omitted if the standard port of the protocol is used.
 * <li><b>application</b> the name of the xma-application. Must not contain any "/".
 * <li>optional <b>component</b> the name of the xma-component inside the xma-application. Must not contain any "/".
 * <li>optional <b>resource</b> relative path of a resource inside a xma-application or xma-component. This may contain "/"s.
 * <li>optional <b>parameters</b> parmeters send to the specified resource. name=value pairs seperated by "&".
 * </li>
 *
 * @author s2877, s3595
 * @version $Id: XMA_URI.java 10829 2013-06-25 16:54:01Z dschwarz $
 */
public class XMA_URI {

    /**
     * hostname is defined as: <br>
     * <tt> server_environment.installation</tt>
     * <p>
     * is the part of the hostname, that referes to
     * the installation environment of the AS.
     * e.g.: .server.lan.at
     */
    private String installation_;

    /**
     * the xma application name
     */
    private String application_;

    /**
     * first part of the hostname,
     * currently this is also the application name
     *
     */
    private String server_;

    /**
     * defines the environment we currently run in.
     * e.g.: devel or prod | ...
     *
     * it is also encoded into the hostname.
     */
    private String environment_;

    /**
     * xma component name
     */
    private String component_;

    /**
     * xma resource name
     */
    private String[] resource_;

    /**
     * request parameters
     */
    private HashMap parameter_=new HashMap();

    /** server port */
    private int port_;

    private static final String DEFAULT_PROTOCOL = Statics.PROTO_HTTP;
    private String protocol_ = DEFAULT_PROTOCOL;

    /** encoding used for URLs */
    private static String encoding = System.getProperty("file.encoding");

    /**
     * Set configuration properties.
     * this will only be used by the boot runtime.
     */
    public static void setProperties( Properties pnew ) {
        String enc = pnew.getProperty("boot.transport.urlencoding");
        if(enc!=null) {
            encoding = enc;
        }
    }

    /**
     * Constructs a XMA_URI from a java.net.URL.
     * It parses its information from the given URL.
     * <p>
     * The java.net.URL may contain escaped characters.
     * This characters are unescaped automatically.
     * Additional slashes inbetween or at the end of an url are removed.<br>
     * e.g. http://server/app//comp -> http://server/app/comp<br>
     * e.g. http://server/app/ -> http://server/app<br>
     * Additional '&' in query strings are removed.<br>
     * e.g. http://server/app?&name=sepp&&a=b& -> http://server/app?name=sepp&a=b<br>
     *
     *
     * @param url the URL to parse the information from.
     * @throws MalformedURLException <ul>
     * <li> if a protocoll different from http or https is specified
     * <li> if no server is specified
     * <li> if a parameter in the query string is illegal
     * </ul>
     */
    public XMA_URI(URL url) throws MalformedURLException {
        if( url == null) throw new IllegalArgumentException();

        try {
            protocol_    = url.getProtocol();
            if(!"http".equalsIgnoreCase(protocol_)&&!"https".equalsIgnoreCase(protocol_)) {
                throw new MalformedURLException(url.toString()+" protocol '"+protocol_+"' not supported");
            }

            server_      = URLDecoder.decode(url.getHost(),encoding);
            if(server_==null||server_.length()==0) {
                throw new MalformedURLException(url.toString()+" server missing");
            }

            port_        = url.getPort();
            String path = url.getPath();

            StringTokenizer tok = new StringTokenizer(path,"/");
            if(tok.hasMoreTokens()) {
               application_ = URLDecoder.decode(tok.nextToken(),encoding);
            }
            if(tok.hasMoreTokens()) {
               component_ = URLDecoder.decode(tok.nextToken(),encoding);
            }
            if(tok.hasMoreTokens()) {
                List list = new ArrayList();
                while(tok.hasMoreTokens()) {
                    String res = URLDecoder.decode(tok.nextToken(),encoding);
                    list.add(res);
                }
                resource_ = new String[list.size()];
                for(int i=0;i<list.size();i++) {
                    resource_[i]=(String)list.get(i);
                }
            }

            String query = url.getQuery();
            if(query!=null) {
                for(StringTokenizer to = new StringTokenizer(query,"&");to.hasMoreTokens();) {
                    String param = to.nextToken();
                    int i = param.indexOf('=');
                    if(i>0&&i<param.length()) {
                        String key = URLDecoder.decode(param.substring(0,i),encoding);
                        String value = URLDecoder.decode(param.substring(i+1,param.length()),encoding);
                        parameter_.put(key,value);
                    } else {
                        throw new MalformedURLException(url.toString()+" url-parameter '"+param+"' malformed");
                    }
                }
            }
        } catch (UnsupportedEncodingException uee) {
            throw new RuntimeException(uee);
        }
    }


    /**
     * Constructs a XMA_URI from a string.
     * It uses {@link #XMA_URI(java.net.URL)} ot parse its information from the given string.
     *
     * @param uri the string to parse the information from.
     * @throws MalformedURLException pased throu from java.net.URL and {@link #XMA_URI(java.net.URL)}.
     */
    public XMA_URI(String uri) throws MalformedURLException {
        this(new URL(uri));
    }

    /**
     * Copy-constructor. The constructed XMA_URI will be an exact copy of the given parameter.
     * @param uri to copy its data from.
     */
    public XMA_URI(XMA_URI uri) {
        copyURI(this, uri);
    }

    /**
     * internal constructor
     */
    private XMA_URI() { }

    /**
     * Get copy of this URI that contains only an base URI to the application root.
     * This is for example: <code>http://host/application</code>
     *
     * @return XMA_URI base URI to application
     */
    public XMA_URI getApplicationURI() {
        XMA_URI newuri = new XMA_URI();
        copyURI(newuri, this);
        newuri.component_ = null;
        newuri.resource_  = null;
        newuri.parameter_.clear();
        return newuri;
    }

    /**
     * copy all atrributes of the input uri <code>from</code>
     * to the destination uri <code>to</code>.
     *
     * @param from copy from
     * @param to   copy to
     */
    public static void copyURI(XMA_URI to, XMA_URI from ) {
        to.protocol_=from.protocol_;
        to.installation_=from.installation_;
        to.environment_=from.environment_;
        to.server_=from.server_;
        to.port_=from.port_;
        to.application_=from.application_;
        to.component_=from.component_;
        if(from.resource_==null) {
            to.resource_=null;
        } else {
            to.resource_=new String[from.resource_.length];
            System.arraycopy(from.resource_,0,to.resource_,0,from.resource_.length);
        }
        to.parameter_=new HashMap(from.parameter_);
    }


    /**
     * Adds a parameter to the parameters of the query string of the url.
     * @param key the name of the parameter
     * @param value the value of the parameter
     */
    public void addParameter(String key,String value) {
        parameter_.put(key,value);
    }

    /**
     * Converts this XMA_URI to a <code>java.net.URL</code> using the http-protocol.
     * @return  URL  a URL required by http transport layer
     */
    public URL getHTTP_URI() {
        return toURL( toString());
    }

    /**
     * Get a String representation of this XMA_URI. This String represents a valid HTTP-URL.
     * Characters considered unsafe for HTTP-URLs are escaped automatically.
     *
     * @return String uri-string
     */
    public String toString() {
        try {
            StringBuffer url = new StringBuffer(32);
            url = getHostApp(url);
            if(component_!=null) url.append("/"+URLEncoder.encode(component_,encoding)); //$NON-NLS-1$
            if(resource_!=null) {
                for(int i=0;i<resource_.length;i++) {
                    url.append("/"+URLEncoder.encode(resource_[i],encoding)); //$NON-NLS-1$
                }
            }
            appendQuery(url);
            return url.toString();
        } catch (UnsupportedEncodingException uee) {
            throw new RuntimeException(uee);
        }

    }


    /**
     * Converts the given String to a <code>java.net.URL</code>. The String
     * must contain a valid url-string.
     * @param url containing a valid url-string.
     * @return URL the converted URL
     */
    private URL toURL( String url ) {
        try {
            return new URL(url);
        } catch (MalformedURLException e) {
            Logger.getLogger( "boot.XMA_URI" ).log(LogLevel.WARNING, "exception:", e);
            throw new RuntimeException(e);
        }
    }

    /**
     * Appends protocol, server and if not empty environment and installation
     * and port to the given StringBuffer.
     * @param url to append
     * @return the given url with the mentioned data added.
     */
    private StringBuffer getHostPort( StringBuffer url ){
        try {
            url.append( protocol_ );
            url.append( "://" + URLEncoder.encode(server_ ,encoding)); //$NON-NLS-1$
            if( environment_!=null  ) url.append(environment_);
            if( installation_!=null ) url.append(installation_);
            if( port_ > 0           ) url.append( ':' + Integer.toString(port_ ) );
            return url;
        } catch (UnsupportedEncodingException uee) {
            throw new RuntimeException(uee);
        }
    }

    /**
     * Appends the application part of the URI to the given StringBuffer
     * @param url to append
     * @return the given url with the mentioned data added.
     */
    private StringBuffer getHostApp( StringBuffer url ){
        try {
            url = getHostPort(url);
            if (application_ != null) {
                url.append("/"); //$NON-NLS-1$
                url.append(URLEncoder.encode(application_,encoding));
            }
            return url;
        } catch (UnsupportedEncodingException uee) {
            throw new RuntimeException(uee);
        }
    }
    
    /**
     * Return start of application url, up to application part
     * @return url string
     */
    public String getHostApp() {
        return getHostApp(new StringBuffer()).toString();
    }

    /**
     * Appends the query parameters of the URI to the given StringBuffer
     * @param url to append
     * @return the given url with the mentioned data added.
     */
    private StringBuffer appendQuery( StringBuffer url ){
        try {
            if(parameter_.size()>0) {
                url.append("?"); //$NON-NLS-1$
                for(Iterator it=parameter_.keySet().iterator();it.hasNext();){
                    String key = (String)it.next();
                    String value = (String)parameter_.get(key);
                    if(key!=null) key = URLEncoder.encode(key,encoding);
                    if(value!=null) value = URLEncoder.encode(value,encoding);
                    url.append(key+"="+value); //$NON-NLS-1$
                    if(it.hasNext()) url.append("&"); //$NON-NLS-1$
                }
            }
            return url;
        } catch (UnsupportedEncodingException uee) {
            throw new RuntimeException(uee);
        }
    }

    /**
     * Gets a HTTP-URL to a resource relative to the application specified
     * by this XMA_URI.
     * Used to specify a custom relative context path within the application.
     *
     * @return  URL  a URL required by http transport layer
     */
    public URL getHTTP_AppUri( String context ) {
        try {
            StringBuffer url = new StringBuffer(32);
            url = getHostApp(url);

            if (context != null) {
                StringTokenizer tok = new StringTokenizer(context,"/");
                while(tok.hasMoreTokens()) {
                    String res = URLEncoder.encode(tok.nextToken(),encoding);
                    url.append( "/" + res ); //$NON-NLS-1$
                }
            }
            return toURL(url.toString());

        } catch (UnsupportedEncodingException uee) {
            throw new RuntimeException(uee);
        }
    }

    /**
     * Get the default protocol used by the transport layer.
     * The default protocol is http.
     * @return the default protocol
     */
    public static String getDEFAULT_PROTOCOL() {
        return DEFAULT_PROTOCOL;
    }

    /**
     * Get the application part of the URI.
     * @return the application part
     */
    public String getApplication() {
        return application_;
    }

    /**
     * Get the component part of the URI
     * @return the component part
     */
    public String getComponent() {
        return component_;
    }

    /**
     * Get the environment part of the URI
     * @return the environment part
     */
    public String getEnvironment() {
        return environment_;
    }

    /**
     * Get the installation part of the URI
     * @return the installation part
     */
    public String getInstallation() {
        return installation_;
    }

    /**
     * Get the port specified in the URI
     * @return the port number
     */
    public int getPort_() {
        return port_;
    }

    /**
     * Set the port for the URI
     * @param port the port to use
     */
    public void setPort(int port) {
        port_ = port;
    }

    /**
     * Get the protocol of the URI
     * @return the protocol
     */
    public String getProtocol_() {
        return protocol_;
    }

    /**
     * Sets the protocl of the URI.
     * Only http and https are supported.
     * @param protocol the protocol to use in this URI
     */
    public void setProtocol(String protocol) {
        if("http".equalsIgnoreCase(protocol_) || "https".equalsIgnoreCase(protocol_)) {
            protocol_=protocol;
        } else {
            throw new RuntimeException("protocol '"+protocol+"' not supported");
        }
    }

    /**
     * Get the resource part of the URI
     * @return the resource part
     */
    public String getResource() {
        if(resource_==null) return null;
        else if(resource_.length==1) return resource_[0];
        else {
            StringBuffer result = new StringBuffer();
            for(int i=0;i<resource_.length-1;i++) {
                result.append(resource_[i]);
                result.append('/');
            }
            result.append(resource_[resource_.length-1]);
            return result.toString();
        }
    }

    /**
     * Get the server part of the URI
     * @return the server part
     */
    public String getServer() {
        return server_;
    }

    /**
     * Set the component part of the URI.
     * @param component the component name relative to this URI.
     * @return this pointer
     */
    public void setComponent(String component) {
        component_ = component;
    }

    /**
     * Set the resource part of the URI
     *
     * @param resource relative to this URI.
     */
    public void setResource(String resource) {
        StringTokenizer tok = new StringTokenizer(resource,"/");
        if(tok.hasMoreTokens()) {
            List list = new ArrayList();
            String res = tok.nextToken();
            list.add(res);
            while(tok.hasMoreTokens()) {
                res = tok.nextToken();
                list.add(res);
            }
            resource_ = new String[list.size()];
            for(int i=0;i<list.size();i++) {
                resource_[i]=(String)list.get(i);
            }
        }
    }

}
